#!/usr/bin/perl -w

#############################################################################
# Template program for accessing ESPRESSO and retrieving prediction results #
#   Written by S. Hirose @CBRC                                              #
#############################################################################

use strict ;
use LWP::UserAgent ;
use LWP::Simple ;

## help message ##
if ( $#ARGV != 0 ) { &help ; exit ; }

## variables ##
my $infile     = $ARGV[0] ;  # input data
my @InputData             ;

my @infileName = split /\./ , $infile         ;
my $outfile    = $infileName[0] . ".ESPRESSO" ;
my $errorfile  = $infileName[0] . ".error"    ;

## file check
if ( -e $outfile )   { print "$outfile exist! Please delete it.\n"   ; exit ; }
if ( -e $errorfile ) { print "$errorfile exist! Please delete it.\n" ; exit ; }

########
# Main #
########
&GetInputData(\$infile,\@InputData) ;  # obtain data for expression/solubility prediction

my $header_check = 0 ;
for (my $i=0; $i<=$#InputData; $i++) {
    print "$InputData[$i]->{'SequenceID'}\n" ;

    my $data_check = &CheckInputData(\$i,\@InputData) ;  # check input data
    
    if ( $data_check >= 1 &&  $data_check < 3 ) {
###	my $pred_result = &access2ESPRESSO(\$i,\@InputData)    ;  # execute ESPRESSO 

        my $pred_result = "" ;

        if ( $data_check == 1 ) {
            $pred_result = &access2ESPRESSO(\$i,\@InputData,\$data_check)    ;  # execute ESPRESSO
        } elsif ( $data_check == 2 ) {
            $pred_result = &access2ESPRESSO(\$i,\@InputData,\$data_check)    ;
        }

	if ( $pred_result eq "" ) {
	    open(SAVE_ERROR,">>$errorfile") ;
	    print SAVE_ERROR "$InputData[$i]->{'SequenceID'}\tCalculation failed. Please recalculate again\n" ; # Error Code
	    close SAVE_ERROR ;   
	} else {
	    
##	    my %result_info    = &parsePredictionResult($pred_result) ;  # get prediction results
	    my %result_info    = &parsePredictionResult($pred_result,$InputData[$i]{'PredictionType'},$InputData[$i]{'ExpressionSystem'}) ;  # get prediction results
	    if ( $InputData[$i]{'PredictionType'} eq "EXP" ) {
		$result_info{'PropPredResult'}  = "Expressed"     if $result_info{'PropPredResult'}  eq "yes" ;
		$result_info{'PropPredResult'}  = "Not Expressed" if $result_info{'PropPredResult'}  eq "no"  ;
		$result_info{'MotifPredResult'} = "Expressed"     if $result_info{'MotifPredResult'} eq "yes" ;
		$result_info{'MotifPredResult'} = "Not Expressed" if $result_info{'MotifPredResult'} eq "no"  ;
		
	    }
	    if ( $InputData[$i]{'PredictionType'} eq "SOL" ) {
		$result_info{'PropPredResult'}  = "Soluble"       if $result_info{'PropPredResult'}  eq "yes" ;
		$result_info{'PropPredResult'}  = "Insoluble"     if $result_info{'PropPredResult'}  eq "no"  ;
		$result_info{'MotifPredResult'} = "Soluble"       if $result_info{'MotifPredResult'} eq "yes" ;
		$result_info{'MotifPredResult'} = "Insoluble"     if $result_info{'MotifPredResult'} eq "no"  ;
		
	    }
	    
	    if (  $header_check == 0 ) {
		open(SAVE_RESULT,">>$outfile") ;
		print SAVE_RESULT "Sequence ID\tSequence\tHost\t"                                      ;
		print SAVE_RESULT "Property based prediction\tProperty based prediction probability\t" ;
		print SAVE_RESULT "Motif based prediction\tPositive Motif(s)\tNegative Motif(s)\t"     ;
		print SAVE_RESULT "Pfam\tSecondary Structure\tDisordered region\tPDB\n"                ;
		close SAVE_RESULT              ;
	    }
	    $header_check = 1 ;
	    
	    open(SAVE_RESULT,">>$outfile") ;
	    print SAVE_RESULT "$InputData[$i]{'SequenceID'}\t"       ;
	    print SAVE_RESULT "$InputData[$i]{'Sequence'}\t"         ;
	    print SAVE_RESULT "$InputData[$i]{'ExpressionSystem'}\t" ;
	    print SAVE_RESULT "$result_info{'PropPredResult'}\t"     ;
	    print SAVE_RESULT "$result_info{'PropPredProb'}\t"       ;
	    print SAVE_RESULT "$result_info{'MotifPredResult'}\t"    ;
	    print SAVE_RESULT "$result_info{'PositiveMotif'}\t"      ;
	    print SAVE_RESULT "$result_info{'NegativeMotif'}\t"      ;
	    print SAVE_RESULT "$result_info{'Pfam'}\t"               ;
	    print SAVE_RESULT "$result_info{'SS'}\t"                 ;
	    print SAVE_RESULT "$result_info{'DISO'}\t"               ;
	    print SAVE_RESULT "$result_info{'PDB'}\n"                ;
	    close SAVE_RESULT ;
	    
	}
    } elsif ( $data_check == 3 ) {
	open(SAVE_ERROR,">>$errorfile") ;
	print SAVE_ERROR "$InputData[$i]->{'SequenceID'}\tNo prediction. This sequence may be expressed\n" ;  # no prediction
	close SAVE_ERROR                ;
    } elsif ( $data_check == 4 ) {
	open(SAVE_ERROR,">>$errorfile") ;
	print SAVE_ERROR "$InputData[$i]->{'SequenceID'}\tESPRESSO does not provide the prediction for protein expression in Brebibacillus expression system\n" ; # no prediction
	close SAVE_ERROR ;
    } else {
	open(SAVE_ERROR,">>$errorfile") ;
	print SAVE_ERROR "$InputData[$i]->{'SequenceID'}\tInput Error\n" ; # Error Code
	close SAVE_ERROR                ;
    }
} 




##############
# Subroutine #
##############
sub help {
print <<EOF;
--------------------------------------------------------------------------------------------
usage : $0  input_file
--------------------------------------------------------------------------------------------
input_file format :
Sequence ID<tab>Sequence Type<tab>Sequence<tab>Prediction Type<tab>Expression System<tab>
********************************************************************************************
Sequence ID: name of your sequence
Sequence Type: "nt" or "AA"
Sequence: nucletide or amino acid sequence 
Prediction Type: "EXP" or "SOL"
Expression System: "Ecoli" or "Wheat"
********************************************************************************************
EOF
}


sub GetInputData {

    my ($file,$data) = (@_) ;
    
    open(OPEN_IN,"<$$file") ;
    while(<OPEN_IN>) {
	chomp ;
	my @input_raw = split /\t/ , $_ ;
	my $hash = {
	    SequenceID       => $input_raw[0] ,
	    SequenceType     => $input_raw[1] ,
	    Sequence         => $input_raw[2] ,
	    PredictionType   => $input_raw[3] ,
	    ExpressionSystem => $input_raw[4] 
	    } ;
	push (@$data,$hash) ;
    }
    close OPEN_IN ;
}


sub CheckInputData {
    my ($num,$data) = (@_) ;

    ## PredictionType ##
    if ( $$data[$$num]->{'PredictionType'} ne "EXP" &&  $$data[$$num]->{'PredictionType'} ne "SOL" ) {
	return (0) ;
    }

    ## ExpressionSystem ##
    if ( $$data[$$num]->{'ExpressionSystem'} ne "Ecoli" && $$data[$$num]->{'ExpressionSystem'} ne "Wheat" && $$data[$$num]->{'ExpressionSystem'} ne "Brevi" ) {
	return (0) ;
    }

    ## Sequence Check ##
    my $check_seq = 0 ;

    # nucleotide
    if ( $$data[$$num]->{'SequenceType'} eq "nt" ) {
	## error single code
	foreach my $nt ( split // , $$data[$$num]->{'Sequence'}) {
	    if ( $nt !~ /[A|C|G|T]/ ) {
		$check_seq = 1 ;
	    }
	}
	## translation
	my $seq_len = length $$data[$$num]->{'Sequence'} ;
	if ( $seq_len % 3 != 0 ) {
	    $check_seq = 1 ;
	}
	## length
	my $seq_length = length ($$data[$$num]->{'Sequence'}) ;
#	if ( $seq_length <= 60 && $seq_length > 3000 ) {
#	    $check_seq = 1 ;
#	}
        if ( $seq_length <= 60 ) {
            $check_seq = 1 ;
        }
        if ( $seq_length > 3000 ) {
            $check_seq = 2 ;
        }
    }

    # amino acid
    if ( $$data[$$num]->{'SequenceType'} eq "AA" ) {
	## error single code
	foreach my $aa ( split // , $$data[$$num]->{'Sequence'}) {
	    if ( $aa !~ /[A|C|D|E|F|G|H|I|K|L|M|N|P|Q|R|S|T|V|W|Y]/ ) {
		$check_seq = 1 ;
	    }
	}
	## length 
	my $seq_length = length ($$data[$$num]->{'Sequence'}) ;
#	if ( $seq_length <= 20 && $seq_length > 1000 ) {
#	    $check_seq = 1 ;
#	}
        if ( $seq_length <= 20 ) {
            $check_seq = 1 ;
        }
        if ( $seq_length > 1000 ) {
            $check_seq = 2 ;
        }
    }


    ## capital letter
    $$data[$$num]->{'Sequence'} =~ tr/[a-z]/[A-Z]/ ;

    if ( $check_seq == 1 ) {
	return (0) ;
    }
	if ( $check_seq == 2 ) {
		return (2) ;
	}


    # No prediction for protein expression in wheat germ cell-free
    if ( $$data[$$num]->{'PredictionType'} eq "EXP" && $$data[$$num]->{'ExpressionSystem'} eq "Wheat" ) {
	return (3) ;
    }
    if ( $$data[$$num]->{'PredictionType'} eq "EXP" && $$data[$$num]->{'ExpressionSystem'} eq "Brevi" ) {
	return (4) ;
    }
   
    return (1) ;
}


sub access2ESPRESSO {
#    my ($num,$data) = (@_) ;
    my ($num,$data,$flag) = (@_) ;
 
   
    ## create jog ID
    my $inputTime = time()                  ;
    my $processId = getppid()               ;
    my $calcID    = $inputTime . $processId ;

    ## access ESPRESSO
    my $ua = new LWP::UserAgent ;
    $ua->agent("AgentName/0.1" . $ua->agent) ;

#    my $req = new HTTP::Request POST => 'http://mbs.cbrc.jp/ESPRESSO/Prediction_S.php' ;
    my $req = "" ;
    if ( $$flag == 1 ) {
#        $req = new HTTP::Request POST => 'http://mbs.cbrc.jp/ESPRESSO/Prediction_S.php' ;
#        $req = new HTTP::Request POST => 'http://cblab.meiyaku.jp/ESPRESSO/Prediction_S.php' ;
        $req = new HTTP::Request POST => 'http://cblab.my-pharm.ac.jp/ESPRESSO/Prediction_S.php' ;
    } elsif ( $$flag == 2 ) {
#        $req = new HTTP::Request POST => 'http://mbs.cbrc.jp/ESPRESSO/Prediction_L.php' ;
#        $req = new HTTP::Request POST => 'http://cblab.meiyaku.jp/ESPRESSO/Prediction_L.php' ;
        $req = new HTTP::Request POST => 'http://cblab.my-pharm.ac.jp/ESPRESSO/Prediction_L.php' ;
    }
    $req -> content_type('application/x-www-form-urlencoded')                          ;
    $req -> content("Model=$$data[$$num]->{'PredictionType'}&System= $$data[$$num]->{'ExpressionSystem'}&Data= $$data[$$num]->{'SequenceType'}&Sequence= $$data[$$num]->{'Sequence'}&calcID=$calcID")                   ;
    my $res = $ua->request($req);

    ## get prediction Result 
    my $tmp_file    = "tmp.result"                                      ;
    my $resultFile  = $calcID . ".result"                               ;
    my $url         = "http://mbs.cbrc.jp/ESPRESSO/RESULTS/$resultFile" ;
    my $CheckResult = getstore($url,$tmp_file)                          ; # Check result file
    my $result      = ""                                                ;
    do {
        system("sleep 10") ;
        $CheckResult = getstore($url,$tmp_file) ;
    } until $CheckResult eq "200" ;
    if ( $CheckResult eq "200" ) {
	$result      = get($url)                                        ;
	unlink $tmp_file                                                ;
    }

    return ($result) ;
}


sub parsePredictionResult {

	my $expSystem = pop ;
	my $predType  = pop ;

    my $prediction  = pop ;
    my %predictInfo = () ;

    my @predict_data = split /\n/ , $prediction ;

    foreach my $info (@predict_data) {
	## property based prediction
	if ( $info =~ /^PROP_PRED/ ) {
	    my @prop_pred_info = split /\t/ , $info ;
	    $predictInfo{'PropPredResult'} = $prop_pred_info[1] ;
	    $predictInfo{'PropPredProb'}   = $prop_pred_info[2] ;
	}

	## motif based prediction
	if ( $info =~ /^MOTIF_PRED/ ) {
	    my @motif_pred_info = split /\t/ , $info , -1 ;
	    $predictInfo{'MotifPredResult'} = $motif_pred_info[1] ;
	    # positive motif
	    if ( $motif_pred_info[2] eq "" ) {
		$predictInfo{'PositiveMotif'} = "-" ;
	    } else {
		$predictInfo{'PositiveMotif'} = &getMotifInfo($motif_pred_info[2],$predType,$expSystem) ;
	    }
	    # negative motif
	    if ( $motif_pred_info[3] eq "" ) {
		$predictInfo{'NegativeMotif'} = "-" ;
	    } else {
		$predictInfo{'NegativeMotif'} = &getMotifInfo($motif_pred_info[3],$predType,$expSystem) ;
	    }	    
	}

	## Pfam motif
	if ( $info =~ /^Pfam/ ) {
	    my @Pfam_info = split /\t/ , $info ;
	    if ( $#Pfam_info == 0 ) {
		$predictInfo{'Pfam'} = "-" ;
	    } else {
		for (my $i=1; $i<$#Pfam_info ; $i=$i+2 ) {
		    $predictInfo{'Pfam'} .= $Pfam_info[$i] . "," . $Pfam_info[$i+1] . ";" ;
		}
		chop $predictInfo{'Pfam'} ;
	    }
	}

	## secondary structure 
	if ( $info =~ /^SS/ ) {
	    my @ss_info = split /\t/ , $info ;
	    $predictInfo{'SS'} = $ss_info[1] ;
	}

	## disordered region
	if ( $info =~ /^DISO/ ) {
	    my @dis_info = split /\t/ , $info ;
	    $predictInfo{'DISO'} = $dis_info[1] ;
	}

	## BLAST
	if ( $info =~ /^PDB/ ) {
	    my @pdb_info = split /\t/ , $info ;
	    if ( $#pdb_info != 0 ) {
		my @pdb_hit  = split /\,/ , $pdb_info[1] ;
		$predictInfo{'PDB'} = $pdb_hit[0] . "," . $pdb_hit[1] . "," . $pdb_hit[2] ;
	    } else {
		$predictInfo{'PDB'} = "-" ;
	    }
	}

    }
    
    return (%predictInfo) ;
}


sub getMotifInfo {

	my $method = $_[1] ;
	my $expressionSys = $_[2] ;

    my @motifList = split /\,/ , $_[0] ; 

    my $motifInfo = "" ;

    foreach my $motif (@motifList) {
	my @locate = split /\-/ , $motif                           ;
	my $start  = $locate[0]                                    ;  # start position of motif
	my $end = $start + 6 ;
	if ( $method eq "SOL" && $expressionSys eq "Ecoli" ) {
		$end    = $end - 1 ;                                   ;  # end position of motif
	}
	$motifInfo .= $locate[1] . "," . $start . "-" . $end . ";" ; 
    }
    chop $motifInfo ;

    return ($motifInfo) ;
}
